def relative_to_project(*components):
    import os

    return os.path.abspath(os.path.join(os.path.dirname(__file__), "..", *components))


def mesh_asset(*components):
    import lacecore

    return lacecore.load_obj(relative_to_project(*components), triangulate=True)


def vitra_mesh():
    result = mesh_asset("examples/vitra/vitra_without_materials_triangulated.obj")
    assert len(result.v) > 100
    return result


def coord_set(a):
    return set(tuple(coords) for coords in a)


def assert_same_vertex_set(a, b):
    assert coord_set(a.v) == coord_set(b.v)


def assert_same_face_set(a, b):
    assert coord_set(a.f) == coord_set(b.f)



"""
Command line for invoking utilities in this library.

Example:

    .. code-block:: sh

        python -m entente.cli \\
            examples/vitra/vitra_without_materials_triangulated.obj \\
            examples/vitra/vitra.pp \\
            examples/vitra/vitra_stretched.obj

"""

import click
from .landmarks.landmark_composite_recipe import DEFAULT_RADIUS


@click.group()
def cli():
    pass


@cli.command()
@click.argument("source_mesh")
@click.argument("landmarks")
@click.argument("target_mesh", nargs=-1, required=True)
@click.option("-o", "--out", help="Output path")
def transfer_landmarks(source_mesh, landmarks, target_mesh, out):
    """
    Transfer landmarks defined in LANDMARKS from SOURCE_MESH to the target
    meshes, which must have vertexwise correspondence.
    """
    import os
    import lacecore
    import meshlab_pickedpoints
    from .landmarks.landmarker import Landmarker

    landmarker = Landmarker.load(source_mesh_path=source_mesh, landmark_path=landmarks)

    for target_mesh_path in target_mesh:
        m = lacecore.load_obj(target_mesh_path, triangulate=True)
        landmarks_on_target_mesh = landmarker.transfer_landmarks_onto(m)
        if out is None:
            filename, _ = os.path.splitext(os.path.basename(target_mesh_path))
            out = filename + ".pp"
        with open(out, "w") as f:
            meshlab_pickedpoints.dump(landmarks_on_target_mesh, f)


@cli.command()
@click.argument("recipe")
@click.argument("output_dir", default="./composite_result")
@click.option("--indicator-radius", default=DEFAULT_RADIUS, show_default=True)
def composite_landmarks(recipe, output_dir, indicator_radius):
    """
    Run the landmark composite recipe in the YAML file RECIPE and write it
    to OUTPUT_DIR.
    """
    import os
    import yaml
    import meshlab_pickedpoints
    from .landmarks.landmark_composite_recipe import LandmarkCompositeRecipe

    recipe_obj = LandmarkCompositeRecipe.load(recipe)

    out_landmarks = os.path.join(output_dir, "landmarks")
    os.makedirs(out_landmarks, exist_ok=True)
    with open("{}.pp".format(out_landmarks), "w") as f:
        meshlab_pickedpoints.dump(recipe_obj.composite_landmarks, f)
    with open("{}.yml".format(out_landmarks), "w") as f:
        yaml.dump(recipe_obj.to_json(), f)

    recipe_obj.write_reprojected_landmarks(
        output_dir=output_dir, radius=indicator_radius
    )


if __name__ == "__main__":  # pragma: no cover

    def set_path():
        """
        Magically add this project to the module path.
        """
        import os
        import sys

        sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

    set_path()
    cli()



import vg


def find_rigid_transform(a, b, compute_scale=False, fail_in_degenerate_cases=True):
    """
    Args:
        a: a Nx3 array of vertex locations
        b: a Nx3 array of vertex locations
        a and b are in correspondence -- we find a transformation such that the first
        point in a will be moved to the location of the first point in b, etc.

    Returns: (R,T) such that a.dot(R) + T ~= b
        R is a 3x3 rotation matrix
        T is a 1x3 translation vector

    Based on Arun et al, "Least-squares fitting of two 3-D point sets," 1987.
    See also Eggert et al, "Estimating 3-D rigid body transformations: a
    comparison of four major algorithms," 1997.

    If compute_scale is True, also computes and returns: (s, R,T) such that s*(R.dot(a))+T ~= b

    In noisy cases, when there is a reflection, this algorithm can fail. In those cases the
    right thing to do is to try a less noise sensitive algorithm like RANSAC. But if you want
    a result anyway, even knowing that it might not be right, set fail_in_degenerate_cases=True.

    """
    import numpy as np

    k = vg.shape.check(locals(), "a", (-1, 3))
    vg.shape.check(locals(), "b", (k, 3))

    a = a.T
    b = b.T

    a_mean = np.mean(a, axis=1)
    b_mean = np.mean(b, axis=1)
    a_centered = a - a_mean.reshape(-1, 1)
    b_centered = b - b_mean.reshape(-1, 1)

    c = a_centered.dot(b_centered.T)
    u, s, v = np.linalg.svd(c, full_matrices=False)
    v = v.T
    R = v.dot(u.T)

    if np.linalg.det(R) < 0:
        if (
            np.any(s == 0) or not fail_in_degenerate_cases
        ):  # This is only valid in the noiseless case; see the paper
            v[:, 2] = -v[:, 2]
            R = v.dot(u.T)
        else:
            raise ValueError(
                "find_rigid_transform found a reflection that it cannot recover from. Try RANSAC or something..."
            )

    if compute_scale:
        scale = np.sum(s) / (np.linalg.norm(a_centered) ** 2)
        T = (b_mean - scale * (R.dot(a_mean))).reshape(1, 3)
        return scale, R.T, T
    else:
        T = (b_mean - R.dot(a_mean)).reshape(1, 3)
        return R.T, T


def find_rigid_rotation(a, b, allow_scaling=False):
    """
    Args:
        a: a Nx3 array of vertex locations
        b: a Nx3 array of vertex locations

    Returns: R such that a.dot(R) ~= b

    See link: http://en.wikipedia.org/wiki/Orthogonal_Procrustes_problem
    """
    import numpy as np

    k = vg.shape.check(locals(), "a", (-1, 3))
    vg.shape.check(locals(), "b", (k, 3))

    a = a.T
    b = b.T

    if a.size == 3:
        cx = np.cross(a.ravel(), b.ravel())
        a = np.hstack([a.reshape(-1, 1), cx.reshape(-1, 1)])
        b = np.hstack([b.reshape(-1, 1), cx.reshape(-1, 1)])

    c = a.dot(b.T)
    u, _, v = np.linalg.svd(c, full_matrices=False)
    v = v.T
    R = v.dot(u.T)

    if np.linalg.det(R) < 0:
        v[:, 2] = -v[:, 2]
        R = v.dot(u.T)

    if allow_scaling:
        scalefactor = np.linalg.norm(b) / np.linalg.norm(a)
        R = R * scalefactor

    return R.T



from setuptools import setup, find_packages

version_info = {}
exec(open("entente/package_version.py").read(), version_info)


def load(filename):
    return open(filename, "rb").read().decode("utf-8")


setup(
    name="entente",
    version=version_info["__version__"],
    description="Work with polygonal meshes which have vertex-wise correspondence",
    long_description=load("README.md"),
    long_description_content_type="text/markdown",
    author="Metabolize",
    author_email="github@paulmelnikow.com",
    url="https://github.com/lace/entente",
    project_urls={
        "Issue Tracker": "https://github.com/lace/entente/issues",
        "Documentation": "https://entente.readthedocs.io/en/stable/",
    },
    packages=find_packages(),
    install_requires=load("requirements.txt"),
    extras_require={"landmarker": load("requirements_landmarker.txt")},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Other Audience",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Artistic Software",
        "Topic :: Multimedia :: Graphics :: 3D Modeling",
        "Topic :: Scientific/Engineering :: Visualization",
    ],
)



import lacecore
from .equality import have_same_topology


def composite_meshes(mesh_paths):
    """
    Create a composite as a vertex-wise average of several meshes in
    correspondence. Faces, groups, and other attributes are loaded from the
    first mesh given.

    Args:
        mesh_paths (list): Paths of the meshes to average.

    Returns:
        lacecore.Mesh: The composite mesh.
    """
    if not len(mesh_paths):
        raise ValueError("Expected at least one mesh path")

    first_mesh_path, remaining_mesh_paths = mesh_paths[0], mesh_paths[1:]

    working_mesh = lacecore.load_obj(first_mesh_path, triangulate=True)
    working_v = working_mesh.v.copy()

    for this_mesh_path in remaining_mesh_paths:
        this_mesh = lacecore.load_obj(this_mesh_path, triangulate=True)
        if not have_same_topology(this_mesh, working_mesh):
            raise ValueError(
                "Expected {} to have the same topology as {}".format(
                    this_mesh_path, first_mesh_path
                )
            )
        working_v += this_mesh.v

    working_v /= len(mesh_paths)

    return lacecore.Mesh(
        v=working_v, f=working_mesh.f, face_groups=working_mesh.face_groups
    )
